#include "parallel.h"

void main()
{
   static int               Virgin = TRUE;
   static ZIP_MAILER       *Node_Mailer;
   static ZIP_MAILER       *Mailer;
   static int               p;
   static int               q;
   static int               P;
   static int               Q;

   int                      m;
   int                      n;
   int                      M;
   int                      N;
   static int               old_M = 0;
   static int               old_N = 0;

   static int               G2_pm;

   long                     Strategy;

   static short             Row_Dis_Type;
   static Cdistrib         *Row_Dis;
   static void            (*Row_Distribution)();
   static int             (*Row_Inverse_Distribution)();
   static int             (*Row_Distribution_limit)();
   static int             (*Row_Distribution_init)();
   static index             Local_Max_Row;
   static int               Row_Extra_param1;
   static int               Row_Extra_param2;
   static void             *Row_Extra;

   static short             Col_Dis_Type;
   static Cdistrib         *Col_Dis;
   static void            (*Col_Distribution)();
   static int             (*Col_Inverse_Distribution)();
   static int             (*Col_Distribution_limit)();
   static int             (*Col_Distribution_init)();
   static index             Local_Max_Col;
   static int               Col_Extra_param1;
   static int               Col_Extra_param2;
   static void             *Col_Extra;

   static matrix_cube_bin   seq_Matrix;
   static CMdistrib        *Matrix_Dis;
   static Dmatrix          *Csparse_Matrix;
   static Smatrix          *Matrix;
   static DXpivot_group    *pivgrp;
   static Dlu_info         *LU_Info;

   static Method           *linear_solver;
   static Extra            *linear_solver_extra;

   static tbx_timer        *Timer;

   int                    i;


   if (Virgin == TRUE)
   {
      Virgin = FALSE;

      if (Zip_init() == NULL)
      {
         print("error - zip_init()");
         exit(1);
      }

      if ((Mailer = g2_grid_open(&P, &Q, NULL)) == NULL)
      {
         print("error - g2_grid_open()");
         exit(1);
      }

      Zip_push(Mailer);
      G2_pq(p, q);
      G2_pm = (!p) && (!q);

      receive_intro_letter(Mailer, &Row_Dis_Type, &Col_Dis_Type,
                           &Row_Extra_param1, &Row_Extra_param2,
                           &Col_Extra_param1, &Col_Extra_param2, &Strategy);
#ifdef DEBUG_main_node
      print("Row_Extra: Par1 = %d, Par2 = %d; Col_Extra: Par1 = %d Par2 = %d",
            Row_Extra_param1, Row_Extra_param2, Col_Extra_param1,
            Col_Extra_param2);
#endif
       /* create the pivoting structure: */
       pivgrp = new_DXpivot_group(Strategy, CSPAR_PIV_IDX_NSCALE);

       seq_Matrix.Array_len = 0;

       Timer = new_tbx_timer();
   }

   do
   {
      receive_matrix_letter(Mailer, &seq_Matrix);

      M = seq_Matrix.M;
      N = seq_Matrix.N;

#ifdef DEBUG_main_node
      print("Matrix size = (%d, %d)\n", seq_Matrix.M, seq_Matrix.N);
#endif

      if ((old_M != M) || (old_N != N))
      {
         if ((old_M != 0) || (old_N != 0))
         {
            /* nuke the structure back to stone age. */

            Free(Matrix_Dis -> rdis -> data);
            Free(Matrix_Dis -> rdis);
            Free(Row_Extra);
            Free(Matrix_Dis -> cdis -> data);
            Free(Matrix_Dis -> cdis);
            Free(Col_Extra);
            Free(Matrix_Dis -> data);
            Free(Matrix_Dis);
            free_Dlu_info(LU_Info);
            Matrix = NULL;
         }

#ifdef DEBUG_main_node
         print("Row_Dis_Type = %hd, Col_Dis_Type = %hd\n", Row_Dis_Type,
               Col_Dis_Type);
#endif

         set_distribution(Mailer, Row_Dis_Type, CDISTRI_ROW, &Row_Dis,
                          &Row_Distribution, &Row_Inverse_Distribution,
                          &Row_Distribution_limit, &Row_Distribution_init,
                          P, p, M, &m, Row_Extra_param1, Row_Extra_param2,
                          &Row_Extra);

#ifdef DEBUG_main_node
         print("node(p=%d,q=%d): parallela -- after row distribution\n", p, q);
#endif

         set_distribution(Mailer, Col_Dis_Type, CDISTRI_COLUMN, &Col_Dis,
                          &Col_Distribution, &Col_Inverse_Distribution,
                          &Col_Distribution_limit, &Col_Distribution_init,
                          Q, q, N, &n, Col_Extra_param1, Col_Extra_param2,
                          &Col_Extra);

#ifdef DEBUG_main_node
         print("node(p=%d,q=%d): parallela -- after col distribution\n", p, q);
#endif

         /* create the matrix distribution structure */
         Matrix_Dis = new_CMdistrib(Row_Dis, Col_Dis);

         /* create the matrix data structure */
         Csparse_Matrix = new_Dmatrix(Matrix_Dis, M, N);

         /* reference the matrix from the structure */
         Matrix = Csparse_Matrix -> Local_Matrix;

         /* create the LU information structure: */
         LU_Info = new_Dlu_info(Csparse_Matrix, pivgrp);

         linear_solver_extra = NULL;
         linear_solver = new_method(linear_system, linear_solver_extra);

         old_M = M;
         old_N = N;
      }
      else
      {
         clear_Smatrix(Matrix);
      }

      /* load the matrices into the structure */

      print("seq_Matrix.num_element = %d\n", seq_Matrix.num_element);

      for (i = 0; i < seq_Matrix.num_element; i++)
      {
#ifdef DEBUG_main_node
         print("(q, q, i, j) = (%hd, %hd, %hd, %hd): Matrix Value = %lf\n",
               p, q, (index) seq_Matrix.i[i], (index) seq_Matrix.j[i],
                seq_Matrix.Value[i]);
#endif
         add_new_value_or_entry(Matrix, (index) seq_Matrix.i[i],
                                (index) seq_Matrix.j[i], seq_Matrix.Value[i]);
      }

      print("Start LU factor\n");
      G2_sync();
      /* start timer */
      tbx_timer_start(Timer);

      CALL_METHOD(linear_solver)(LU_Info, NULL, NULL, XTRA(linear_solver));

      /* stop timer */
      tbx_timer_stop(Timer);

      /* print time */
      compute_time(Timer);

#ifdef DEBUG_main_node
      print_time(Timer);
#endif

      transmit_timing_letter(Mailer, Timer);
      tbx_timer_reset(Timer);

   } while (1);

#if 0
   Zip_pop();
   zip_exit();
#endif
}
